import sys
import numpy as np
from numpy import where, mean, zeros, sqrt, atleast_1d, asarray
from numpy.linalg import eigvalsh
from scipy.stats import chi2
from chi2comb import chi2comb_cdf, ChiSquared


def davies_pvalue(q, w):
    r"""Joint significance of statistics derived from chi2-squared distributions.

    Parameters
    ----------
    q : float
        Test statistics.
    w : array_like
        Weights of the linear combination.

    Returns
    -------
    float
        Estimated p-value.
    """

    q = asarray(atleast_1d(q), float)
    w = asarray(w, float)

    re = _pvalue_lambda(_lambda(w), q)
    param = dict()
    param["liu_pval"] = re["p_val_liu"][0]
    param["Is_Converged"] = re["is_converge"][0]

    # p_value_resampling = None

    # re = dict(
    #     p_value=re["p_value"][0],
    #     param=param,
    #     p_value_resampling=p_value_resampling,
    #     pval_zero_msg=re["pval_zero_msg"],
    # )
    return re["p_value"][0]


def _pvalue_lambda(lambda_, Q):

    n1 = len(Q)

    p_val = zeros(n1)
    p_val_liu = zeros(n1)
    is_converge = zeros(n1)

    p_val_liu = _liu_pvalue_mod_lambda(Q, lambda_)

    for i in range(n1):
        chi2s = [ChiSquared(w, 0., 1) for w in lambda_]
        out = chi2comb_cdf(Q[i], chi2s, 0., lim=10000, atol=10 ** -6)

        p_val[i] = 1 - out[0]

        is_converge[i] = 1

        # check convergence
        if len(lambda_) == 1:
            p_val[i] = p_val_liu[i]
        elif out[1] != 0:
            is_converge[i] = 0

        # check p-value
        if p_val[i] > 1 or p_val[i] <= 0:
            is_converge[i] = 0
            p_val[i] = p_val_liu[i]

    p_val_msg = None
    p_val_log = None
    if p_val[0] == 0:
        # TODO: the tests have never been here. Come up with a tests that arrives here.
        param = _liu_params_mod_lambda(lambda_)
        p_val_msg = _liu_pvalue_mod_lambda_zero(
            Q[0],
            param["muQ"],
            param["muX"],
            param["sigmaQ"],
            param["sigmaX"],
            param["ll"],
            param["d"],
        )
        p_val_log = _liu_pvalue_mod_lambda(Q[0], lambda_, log_p=True)[0]

    return dict(
        p_value=p_val,
        p_val_liu=p_val_liu,
        is_converge=is_converge,
        p_val_log=p_val_log,
        pval_zero_msg=p_val_msg,
    )


def _lambda(K):

    lambda1 = eigvalsh(K)
    lambda1 = np.sort(lambda1)
    idx1 = where(lambda1 >= 0)[0]

    # eigenvalue bigger than sum(eigenvalues)/1000
    idx2 = where(lambda1 > mean(lambda1[idx1]) / 100000)[0]

    if len(idx2) == 0:
        sys.exit("No Eigenvalue is bigger than 0!!")
    return lambda1[idx2]


def _liu_pvalue_mod_lambda(Q_all, lambda_, log_p=False):

    param = _liu_params_mod_lambda(lambda_)

    Q_Norm = (Q_all - param["muQ"]) / param["sigmaQ"]
    Q_Norm1 = Q_Norm * param["sigmaX"] + param["muX"]

    if log_p:
        assert False
        # Q_Norm1 = exp(Q_Norm1)
    p_value = 1 - chi2.cdf(Q_Norm1, df=param["ll"], loc=param["d"])

    return p_value


def _liu_params_mod_lambda(lambda_):
    # Helper function for getting the parameters for the null approximation

    c1 = zeros(4)
    for i in range(4):
        c1[i] = sum(lambda_ ** (i + 1))

    muQ = c1[0]
    sigmaQ = sqrt(2 * c1[1])
    s1 = c1[2] / c1[1] ** (3 / 2)
    s2 = c1[3] / c1[1] ** 2

    if s1 ** 2 > s2:
        a = 1 / (s1 - sqrt(s1 ** 2 - s2))
        d = s1 * a ** 3 - a ** 2
        ll = a ** 2 - 2 * d
    else:
        ll = 1. / s2
        a = sqrt(ll)
        d = 0

    muX = ll + d
    sigmaX = sqrt(2) * a

    return dict(ll=ll, d=d, muQ=muQ, muX=muX, sigmaQ=sigmaQ, sigmaX=sigmaX)


def _liu_pvalue_mod_lambda_zero(Q, muQ, muX, sigmaQ, sigmaX, l, d):

    assert False
    # TODO: Unit tests are not currently testing this function.
    # temp = c(
    #     0.05,
    #     10 ** -10,
    #     10 ** -20,
    #     10 ** -30,
    #     10 ** -40,
    #     10 ** -50,
    #     10 ** -60,
    #     10 ** -70,
    #     10 ** -80,
    #     10 ** -90,
    #     10 ** -100,
    # )

    # out = qchisq(temp, df=l, ncp=d, lower_tail=False)

    # IDX = max(where(out < Q.Norm1))

    # pval_msg = "Pvalue < %e" % temp[IDX]
    # return pval_msg



from chiscore import optimal_davies_pvalue, data_file
from numpy import load
from numpy.testing import assert_allclose


def test_optimal_davies_pvalue():
    with data_file("optimal_davies_pvalue.npz") as filepath:
        data = load(filepath)

    pval = optimal_davies_pvalue(*data["args"])
    assert_allclose(pval, data["pval"])


def main():
    q = [1.5, 3.0]
    mu = -0.5
    var = 1.0
    kur = 3.0
    w = [10.0, 0.2, 0.1, 0.3]
    remain_var = 0.5
    df = 3.4
    trho = [5.1, 0.2]
    grid = [0., 0.01]
    print(optimal_davies_pvalue(q, mu, var, kur, w, remain_var, df, trho, grid))


if __name__ == "__main__":
    main()



from chiscore import davies_pvalue, data_file
from numpy import load
from numpy.testing import assert_allclose


def test_davies_pvalue():
    with data_file("davies_pvalue.npz") as filepath:
        data = load(filepath)

    assert_allclose(davies_pvalue(*data["args"]), data["pval"])


def main():
    q = 1.5
    w = [[0.3, 5.0], [5.0, 1.5]]
    davies_pvalue(q, w)


if __name__ == "__main__":
    main()



from numpy import sqrt, asarray, sum, isfinite, all, atleast_1d
from scipy.stats import chi2


def mod_liu(q, w):
    r"""Joint significance of statistics derived from chi2-squared distributions.

    Parameters
    ----------
    q : float
        Test statistics.
    w : array_like
        Weights of the linear combination.

    Returns
    -------
    float
        Estimated p-value.
    """

    q = asarray(q, float)
    if not all(isfinite(atleast_1d(q))):
        raise ValueError("There are non-finite values in `q`.")

    w = asarray(w, float)
    if not all(isfinite(atleast_1d(w))):
        raise ValueError("There are non-finite values in `w`.")

    d = sum(w)
    w /= d

    c1 = sum(w)

    c2 = sum(w ** 2)

    c3 = sum(w ** 3)

    c4 = sum(w ** 4)

    s1 = c3 / (c2 ** (3 / 2))

    s2 = c4 / c2 ** 2

    muQ = c1

    sigmaQ = sqrt(2 * c2)

    if s1 ** 2 > s2:

        a = 1 / (s1 - sqrt(s1 ** 2 - s2))

        delta = s1 * a ** 3 - a ** 2

        l = a ** 2 - 2 * delta
        if l < 0:
            raise RuntimeError("This term cannot be negative.")

    else:

        delta = 0
        l = 1 / s2
        a = sqrt(l)

    Q_norm = (q / d - muQ) / sigmaQ * sqrt(2 * l) + l

    Qq = atleast_1d(chi2(df=l).sf(Q_norm))[0]

    return (Qq, muQ * d, sigmaQ * d, l)



from numpy import sqrt, full, inf, asarray, divide, min
from scipy.stats import chi2
from scipy.integrate import quad
from chi2comb import chi2comb_cdf, ChiSquared

_EPSABS = 1e-12


def optimal_davies_pvalue(q, mu, var, kur, w, remain_var, df, trho, grid, pmin=None):
    r"""Joint significance of statistics derived from chi2-squared distributions.

    Parameters
    ----------
    q : array_like
        Most significant of the independent test statistics, qₘᵢₙ(𝜌ᵥ) [1].
    mu : float
        Mean of the linear combination of the chi-squared distributions.
    var : float
        Variance of the linear combination of the chi-squared distributions.
    kur : float
        Kurtosis of the linear combination of the chi-squared distributions.
    w : array_like
        Weights of the linear combination.
    remain_var : float
        Remaining variance assigned to a Normal distribution.
    df : float
        Overall degrees of freedom.
    trho : array_like
        Weight between the combination of chi-squared distributions and independent
        chi-squared distributions.
    grid : array_like
        Grid parameters.
    pmin : float
        Boundary of the possible final p-values.

    Returns
    -------
    float
        Estimated p-value.

    References
    ----------
    [1] Lee, Seunggeun, Michael C. Wu, and Xihong Lin. "Optimal tests for rare variant
    effects in sequencing association studies." Biostatistics 13.4 (2012): 762-775.
    """
    q = asarray(q, float)
    mu = float(mu)
    var = float(var)
    kur = float(kur)

    w = asarray(w, float)
    remain_var = float(remain_var)
    df = float(df)
    trho = asarray(trho, float)
    grid = asarray(grid, float)

    args = (q, mu, var, kur, w, remain_var, df, trho, grid)
    re = quad(_davies_function, 0, 40, args, limit=1000, epsabs=_EPSABS, full_output=1)

    # Might want to add this back in
    if re[1] > 1e-6:
        re = _skat_liu_pvalue(q, mu, var, kur, w, remain_var, df, trho, grid, pmin)
        return re

    pvalue = 1 - re[0]
    if pmin is not None:
        if pmin * len(grid) < pvalue:
            pvalue = pmin * len(grid)
    return pvalue


def _skat_liu_pvalue(
    pmin_q, MuQ, VarQ, KerQ, lambda_, VarRemain, Df, tau, r_all, pmin=None
):
    args = (pmin_q, MuQ, VarQ, KerQ, lambda_, VarRemain, Df, tau, r_all)
    re = quad(_skat_liu_func, 0, 40, args, limit=2000, epsabs=_EPSABS, full_output=1)

    pvalue = 1 - re[0]

    if pmin is not None:
        if pmin * len(r_all) < pvalue:
            pvalue = pmin * len(r_all)

    return pvalue


def _davies_function(x, pmin_q, MuQ, VarQ, KerQ, lambda_, VarRemain, Df, tau, r_all):
    temp1 = tau * x

    temp = divide(
        pmin_q - temp1, 1 - r_all, out=full(temp1.shape, inf), where=r_all != 1.0
    )
    temp_min = min(temp)

    re = 0
    min1 = temp_min
    if min1 > sum(lambda_) * 10 ** 4:
        temp = 0
    else:
        min1_temp = min1 - MuQ
        sd1 = sqrt(VarQ - VarRemain) / sqrt(VarQ)
        min1_st = min1_temp * sd1 + MuQ

        chi2s = [ChiSquared(w, 0., 1) for w in lambda_]
        dav_re = chi2comb_cdf(min1_st, chi2s, 0., lim=10000, atol=10 ** -5)

        temp = 1 - dav_re[0]
        if dav_re[1] != 0:
            msg = "Could not estimate the cdf value: {}".format(str(dav_re))
            raise RuntimeError(msg)
    if temp > 1:
        temp = 1
    re = (1 - temp) * chi2(df=1).pdf(x)
    return re


def _skat_liu_func(x, pmin_q, MuQ, VarQ, KerQ, lambda_, VarRemain, Df, tau, r_all):

    temp1 = tau * x

    temp = (pmin_q - temp1) / (1 - r_all)
    temp_min = min(temp)

    temp_q = (temp_min - MuQ) / sqrt(VarQ) * sqrt(2 * Df) + Df

    return chi2(df=Df).cdf(temp_q) * chi2(df=1).pdf(x)



from os.path import dirname, realpath, join
import tempfile
import shutil
import warnings

_filenames = ["davies_pvalue.npz", "optimal_davies_pvalue.npz", "mod_liu.npz"]


class data_file(object):
    def __init__(self, filenames):
        self._unlist = False
        if not isinstance(filenames, (tuple, list)):
            filenames = [filenames]
            self._unlist = True

        for fn in filenames:
            if fn not in _filenames:
                raise ValueError(
                    "Unrecognized file name {}. Choose one of these: {}".format(
                        fn, _filenames
                    )
                )

        self._dirpath = tempfile.mkdtemp()
        self._filenames = filenames

    def __enter__(self):
        import pkg_resources

        filepaths = [join(self._dirpath, fn) for fn in self._filenames]

        for fn, fp in zip(self._filenames, filepaths):
            if __name__ == "__main__":
                shutil.copy(join(dirname(realpath(__file__)), fn), fp)
            else:
                resource_path = "_data/{}".format(fn)
                content = pkg_resources.resource_string(
                    __name__.split(".")[0], resource_path
                )

                with open(fp, "wb") as f:
                    f.write(content)

        if self._unlist:
            return filepaths[0]
        return filepaths

    def __exit__(self, *_):
        try:
            shutil.rmtree(self._dirpath)
        except PermissionError as e:
            warnings.warn(str(e) + "\n. I will ignore it and proceed.")



"""
Chiscore package
================

Estimate the joint significance of test statistics derived from linear combination
of chi-squared distributions.

Functions
---------
davies_pvalue
optimal_davies_pvalue
mod_liu

References
----------
[1] Lee, Seunggeun, Michael C. Wu, and Xihong Lin. "Optimal tests for rare variant
effects in sequencing association studies." Biostatistics 13.4 (2012): 762-775.
"""

from __future__ import absolute_import

from ._testit import test
from ._davies import davies_pvalue
from ._optimal import optimal_davies_pvalue
from ._liu import mod_liu
from ._data import data_file

__version__ = "0.0.11"

__all__ = [
    "__version__",
    "test",
    "davies_pvalue",
    "optimal_davies_pvalue",
    "mod_liu",
    "data_file",
]
